# -*- coding: utf-8 -*-

# Copyright (C) 2017 github.com/shyal
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import getpass
import vulcan.lib.highlighter as hi

class CardContext(object):

    def __init__(self, card, command, banner, prompt):
        self.card = card
        self.preamble = hi.highlight(card.preamble)
        self.__command = command
        # this is, for example, the python banner, or sqlite3 banner
        # anything you'd have printed when first entering an interactive shell
        self.__banner = banner
        # this is for example >>> in python
        self.__prompt = prompt

    @property
    def prompt(self):
        if isinstance(self.__prompt, str):
            return [self.__prompt]
        elif isinstance(self.__prompt, list):
            return self.__prompt

    @property
    def command(self):
        if isinstance(self.__command, str):
            if self.__command:
                return [self.__command]
            return []
        elif isinstance(self.__command, list):
            return self.__command

    def render(self):
        rendered_context = []
        if self.__banner:
            rendered_context.append(self.__banner)
            rendered_context.append("\n")
        rendered_context.extend(self.prompt)
        rendered_context.extend(self.preamble[:])
        for i, t in enumerate(rendered_context):
            if isinstance(t, tuple):
                if '\n' in t[1]:
                    rendered_context[i] = ('', t[1].replace('\n', '\n'+self.__prompt))
        rendered_context.append("\n")
        rendered_context.extend(self.prompt)
        rendered_context.append("\n")
        rendered_context.extend(self.prompt)
        rendered_context.append(('dark blue', "# "))
        rendered_context.append(('dark blue', self.card.question))
        return rendered_context

class PythonCardContext(CardContext):

    def __init__(self, *args, **kwargs):
        command = "python"
        banner = """Python 2.7.10 (default, Jul 14 2015, 19:46:27)
[GCC 4.2.1 Compatible Apple LLVM 6.0 (clang-600.0.39)] on darwin
Type "help", "copyright", "credits" or "license" for more information."""
        prompt = ">>> "
        super(PythonCardContext, self).__init__(command=command, banner=banner, prompt=prompt, *args, **kwargs)

class SQLite3CardContext(CardContext):

    def __init__(self, *args, **kwargs):
        command = "sqlite3"
        banner = """SQLite version 3.8.7.4 2014-12-09 01:34:36
Enter ".help" for usage hints.
Connected to a transient in-memory database.
Use ".open FILENAME" to reopen on a persistent database."""
        prompt = "sqlite> "
        super(SQLite3CardContext, self).__init__(command=command, banner=banner, prompt=prompt, *args, **kwargs)

class NodeCardContext(CardContext):

    def __init__(self, *args, **kwargs):
        command = "node"
        prompt = "> "
        super(NodeCardContext, self).__init__(command=command, banner='', prompt=prompt, *args, **kwargs)

class RubyCardContext(CardContext):

    def __init__(self, *args, **kwargs):
        command = "ruby"
        banner = """ruby 2.0.0p481 (2014-05-08 revision 45883) [universal.x86_64-darwin14]"""
        prompt = ""
        super(RubyCardContext, self).__init__(command=command, banner=banner, prompt=prompt, *args, **kwargs)

class SQLCardContext(CardContext):

    def __init__(self, *args, **kwargs):
        command = "mysql"
        banner = """Welcome to the MariaDB monitor.  Commands end with ; or \\g.
Your MariaDB connection id is 5
Server version: 10.1.23-MariaDB Homebrew

Copyright (c) 2000, 2017, Oracle, MariaDB Corporation Ab and others.

Type 'help;' or '\\h' for help. Type '\\c' to clear the current input statement."""
        prompt = "MariaDB [(none)]> "
        super(SQLCardContext, self).__init__(command=command, banner=banner, prompt=prompt, *args, **kwargs)

def bash_prompt():
    return [("dark blue", getpass.getuser()), "@", ("dark green", "vulcan"), ":",("dark blue", "~ ")]

class BashCardContext(CardContext):

    def __init__(self, *args, **kwargs):
        super(BashCardContext, self).__init__(command='', banner='', prompt=bash_prompt(), *args, **kwargs)

    def render(self):
        pre = []
        pre.append("\n")
        pre.extend(self.preamble[:])
        pre.append("\n")
        pre.append("\n")
        return pre + bash_prompt() + ["# " + self.card.question]

def prompt(card):
    if card.context == "python":
        return PythonCardContext(card)
    if card.context == "sqlite3":
        return SQLite3CardContext(card)
    if card.context in ['node', 'nodejs', 'react', 'redux']:
        return NodeCardContext(card)
    elif card.context == "sql":
        return SQLCardContext(card)
    elif card.context == "ruby":
        return RubyCardContext(card)
    return BashCardContext(card)



from sqlalchemy import *
from migrate import *
from migrate.changeset import *

meta = MetaData()

def upgrade(migrate_engine):
    meta = MetaData(bind=migrate_engine)
    cards = Table('cards', meta, autoload=True)
    cards.c.last_answered.drop()
    cards.c.closing.drop()
    cards.c.times_asked.drop()
    cards.c.num_successes.drop()

def downgrade(migrate_engine):
    raise NotImplementedError



from urwid import html_fragment
from textwrap import dedent

from controller import VulcanController
import vulcan.lib.store
import os
from collections import namedtuple

def blank_doc():
    return dedent("""
    <style type="text/css">
        pre, span {
            margin-top: -1px;
            margin-bottom: -1px;
            display: inline-block;
        }
        .urwid_fragment_usage {
            font-size: 0.8em;
        }
    </style>
    <div>
    %s
    </div>
    """)

def section(text, screenshot):
    return dedent(f"""
        <div>
            <p>{text}</p>
            <div  class='urwid_fragment_usage'>
            {screenshot}
            </div>
        </div>
    """)

def main(db_path):

    DocEntry = namedtuple("DocEntry", ["text", "key_sequence"])

    entries = [
        DocEntry("Press ctrl n to enter the edit card screen", ['ctrl n']),
        DocEntry("Add a question and an answer", ['down'] + list("Einstein's Famous Equation") + ['down'] + list("e=mc^2") + ['down'] + list("physics") + ['down']),
        DocEntry("Pressing 'Add' adds to card to the deck, and resets the input fields", ['enter']),
        DocEntry("Let's add a second card", ['up']*4 + list("Famous vulcan saying") + ['down'] + list("live long and ________") + ['down'] + list("prosper") + ['down'] + list('trivia') + ['down']),
        DocEntry("And a third", ['enter'] + ['up']*3 + list("Logic is the _______ of our civilization, with which we ascend from chaos, using reason as our guide.") + ['down'] + list("cement") + ['down'] + list('trivia') + ['down']),
        DocEntry("Vulcan now proceeds to asking you those questions", ['enter'] + ['down'] + ['enter']),
        DocEntry("What happens if we enter the wrong answer", list('f = ma') + ['enter']),
        DocEntry("Vulcan has an in-built help menu you can access with F1", ['f1']),
        DocEntry("Let's take a look at your stats by pressing F5", ['f5']),
        DocEntry("Vulcan can also display your cards in a grid by pressing F3", ['f3']),
    ]

    large = (110, 30)
    mid = (80, 20)

    html_fragment.screenshot_init([large], [x.key_sequence for x in entries])

    controller = VulcanController(db_path)
    controller.main(screenshot_mode=True)

    html = html_fragment.screenshot_collect()

    text = ""

    for i, doc in enumerate(html):
        try:
            text += section(entries[i-1].text if i > 0 else '', doc.replace('e5e5e5', '02233f').replace('0000ee', '1f7ad8'))
        except:
            pass

    with open('../docs/_static/usage.html', 'w') as w:
        w.write(blank_doc()%text)



# -*- coding: utf-8 -*-

# Copyright (C) 2017 github.com/shyal
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import re
import json
from base64 import b64decode, b64encode
from textwrap import dedent

from pygments import highlight as pyhighlight
from pygments.formatter import Formatter
from pygments.lexers import PythonLexer

class URWIDFormatter(Formatter):

    def __init__(self, **options):
        Formatter.__init__(self, **options)

        # create a dict of (start, end) tuples that wrap the
        # value of a token so that we can use it in the format
        # method later
        self.styles = {}

        # we iterate over the `_styles` attribute of a style item
        # that contains the parsed style values.
        for token, style in self.style:
            start = end = ''
            # a style item is a tuple in the following form:
            # colors are readily specified in hex: 'RRGGBB'
            if style['color']:
                start += '["%s", "'%self.urwid_cols(style['color'])
                end = '"],' + end
            else:
                start += '["", "'
                end = '"],' + end
            self.styles[token] = (start, end)

    def urwid_cols(self, col):
        urwid_cols = {
            '008000': 'dark blue',
            '888888': 'grey',
            'BA2121': 'dark green',
            '0044DD': 'dark green',
            '666666': 'dark green',
            '408080': 'grey'
        }

        if col in urwid_cols:
            return urwid_cols[col]

        return 'grey'

    def format(self, tokensource, outfile):
        # lastval is a string we use for caching
        # because it's possible that an lexer yields a number
        # of consecutive tokens with the same token type.
        # to minimize the size of the generated html markup we
        # try to join the values of same-type tokens here
        lastval = ''
        lasttype = None

        for ttype, value in tokensource:
            # if the token type doesn't exist in the stylemap
            # we try it with the parent of the token type
            # eg: parent of Token.Literal.String.Double is
            # Token.Literal.String
            while ttype not in self.styles:
                ttype = ttype.parent
            if ttype == lasttype:
                # the current token type is the same of the last
                # iteration. cache it
                lastval += value
            else:
                # not the same token as last iteration, but we
                # have some data in the buffer. wrap it with the
                # defined style and write it to the output file
                if lastval:
                    stylebegin, styleend = self.styles[lasttype]
                    lastval = b64encode(str.encode(lastval)).decode('utf-8')
                    outfile.write(stylebegin + lastval + styleend)
                # set lastval/lasttype to current values
                lastval = value
                lasttype = ttype

        # if something is left in the buffer, write it to the
        # output file, then close the opened <pre> tag
        if lastval:
            stylebegin, styleend = self.styles[lasttype]
            lastval = b64encode(str.encode(lastval)).decode('utf-8')
            outfile.write(stylebegin + lastval + styleend)

class PyGToURWID:

    def __init__(self, code):
        self.code = code
        self.py_highlighted = pyhighlight(code, PythonLexer(), URWIDFormatter())
        r = "["+self.py_highlighted[:-1]+"]"
        j = json.loads(r)
        for i, it in enumerate(j):
            if type(it) is list:
                j[i] = (it[0], b64decode(it[1]).decode('utf-8'))
            else:
                j[i] = b64decode(j[i]).decode('utf-8')

        self.final = j

def highlight(code):
    pygt = PyGToURWID(code)
    return pygt.final

if __name__ == "__main__":

    from urwid import *
    palette = [
        ('dark blue', 'dark blue', '', 'standout'),
        ('dark green', 'dark green', '', 'standout'),
        ('grey', 'light gray', '', 'standout'),
        ('brown', 'brown', '', 'standout'),
    ]

    code = """
    class PyGToURWID:

        def __init__(self, code):
            self.code = code
            self.py_highlighted = highlight(code, PythonLexer(), OldHtmlFormatter())
            r = "["+self.py_highlighted[:-1]+"]"
            j = json.loads(r)
            for i, it in enumerate(j):
                if type(it) is list:
                    j[i] = (it[0], b64decode(it[1]).decode('utf-8'))
                else:
                    j[i] = b64decode(j[i]).decode('utf-8')

            print(j)

            self.final = j
    """
    pygt = PyGToURWID(code)

    loop = MainLoop(Filler(Text(pygt.final), 'top'), palette=palette)
    loop.run()






# -*- coding: utf-8 -*-

# Copyright (C) 2019 github.com/shyal
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import sys
from collections import OrderedDict
from traceback import print_exc

from vulcan.lib.api import FlashcardAlgorithm
from vulcan.lib.store import Cards, Card
from vulcan.lib.prefs import prefs, reset_prefs
from vulcan.lib.store import get_session

class VulcanModel(object):

    def __init__(self, db_path):
        self.session = get_session(db_path)
        self.refresh_tags()
        self.__current_card = None

    def refresh_tags(self):
        self.tags = OrderedDict()
        for tag in sorted(Cards.tags()):
            try:
                self.tags[tag] = prefs.get("tag_"+tag, True)
            except:
                # shelve sucks
                pass
        self.init_algo()

    @property
    def current_tags(self):
        return [k for k, v in self.tags.items() if v]

    def init_algo(self):
        self.algo = FlashcardAlgorithm(wdist_weight=1, tdist_weight=0, includeMuli=True, includePaused=False, tags=self.current_tags)

    def draw_card(self):
        self.algo.init()
        self.__current_card = self.algo.draw_card()
        return self.__current_card

    @property
    def current_card(self):
        return self.__current_card

    @current_card.setter
    def current_card(self, value):
        self.__current_card = value

    def remove_card(self, card):
        if card == self.__current_card:
            self.current_card = None
        self.session.delete(card)
        self.session.commit()

    def add_card(self, question, answer, context, preamble):
        card = Card(question=question, answer=answer, context=context, paused=False, preamble=preamble)
        self.session.add(card)
        self.session.commit()
        self.refresh_tags()

    def pause_card(self, card):
        card.paused = True
        self.session.commit()


# -*- coding: utf-8 -*-

# Copyright (C) 2017 github.com/shyal
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import os
from datetime import datetime
from textwrap import dedent
import arrow
from datetime import datetime

from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import ForeignKey
from sqlalchemy.orm import relationship
from sqlalchemy import Column, Integer, String, VARCHAR, TEXT
import sqlalchemy

_sessions = {}

def get_session(db_path = "~/.vulcan/vulcan.db"):
    path = os.path.expanduser(db_path)
    if 'session' in _sessions:
        return _sessions['session']

    dirname = os.path.dirname(path)
    
    if dirname and not os.path.isdir(dirname):
      os.makedirs(dirname)

    storage_engine = 'sqlite:///{0}'.format(path)

    engine = sqlalchemy.create_engine(storage_engine, echo=False)
    Session = sessionmaker(bind=engine)

    Base.metadata.create_all(engine)
    session = Session()
    _sessions['session'] = session
    return session

Base = declarative_base()

class Common(object):
    def as_dict(self):
        obj = {}
        for k,v in self.__dict__.iteritems():
            if k not in ['_sa_instance_state']:
                obj[k] = v if v != "0" else ''
        return obj

class Card(Base, Common):
    __tablename__ = 'cards'
    id = Column(Integer, primary_key=True)
    question = Column(String, default="")
    answer = Column(String, default="")
    last_answered = Column(Integer, default=0)
    paused = Column(sqlalchemy.types.Boolean, default=False)
    multi = Column(sqlalchemy.types.Boolean, default=False)
    preamble = Column(String, default="")
    closing = Column(String, default="")
    context = Column(String, default="")
    times_asked = Column(Integer, default=0)
    num_successes = Column(Integer, default=0)

    def __repr__(self):
        return dedent("""
            Q: %s
            A: %s
            Successes: %i
            Wrong: %i
            multi: %i
            """) % (self.question, self.answer, self.successes, self.failures, self.multi)

    def print_stats(self, stdout):
        stdout.write("\nSuccesses: %i\nWrong: %i\n"% (self.successes, self.failures))
        stdout.flush()

    @property
    def successes(self):
        return len(list(filter(lambda x: x.success, self.history)))

    @property
    def failures(self):
        return len(list(filter(lambda x: not x.success, self.history)))

    @property
    def attempts(self):
        return len(self.history)

    def pop(self):
        if len(self.history):
            self.num_successes += -self.history[-1]
            self.times_asked -= 1
            self.history = self.history[:-2]

    def as_dict(self):
        card_dict = super(Card, self).as_dict()
        card_dict["history"] = [x.as_dict() for x in self.history]
        return card_dict

    def viewed_today(self):
        for h in self.history:
            if arrow.get(h.date).date() == datetime.now().date():
                return True
        return False

    @property
    def last_draw_timestamp(self):
        return datetime.fromtimestamp(self.last_answered)

    def compare_answer(self, answer):
        if self.context == 'python':
            try:
                from yapf.yapflib.yapf_api import FormatCode
                return FormatCode(self.answer)[0].strip() == FormatCode(answer)[0].strip()
            except:
                pass
        return self.answer.lower().strip() == answer.lower().strip()

class Cards(object):
    
    @staticmethod
    def tags():
        session = get_session()
        return [x[0] for x in session.query(Card.context).distinct()]


class History(Base, Common):
    __tablename__ = 'history'
    id = Column(Integer, primary_key=True)
    date = Column(Integer)
    success = Column(sqlalchemy.types.Boolean)
    qid = Column(Integer, ForeignKey('cards.id'))
    card = relationship("Card", back_populates="history")


Card.history = relationship("History", order_by=History.date, back_populates="card", cascade="save-update, merge, delete")



#!/usr/bin/env python
from migrate.versioning.shell import main
import os

if __name__ == '__main__':

    db_path = 'sqlite:///'+os.path.expanduser('~/.vulcan/vulcan.db')

    main(url=db_path,
         debug='False',
         repository='vulcan_db'
         )



# -*- coding: utf-8 -*-

# Copyright (C) 2019 github.com/shyal
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


from vulcan.lib.store import get_session
from vulcan.lib.store import Card
import json
import os

def stats(today=False, total=False):
    session = get_session()
    r = ""
    if today:
        r += f"reviews today: {get_num_cards_studied_today()}"
    if total:
        r += "total cards: {0}".format(session.query(Card).count())
    return r

def get_num_cards_studied_today():
    session = get_session()
    query = session.query(Card)
    n = 0
    for card in query:
        n += card.viewed_today()
    return n

def export(db_path, json_path):
    session = get_session(db_path)
    query = session.query(Card)
    cards = []
    for card in query:
        cards.append(card.as_dict())
    json.dump(cards, open(os.path.expanduser(json_path), 'w'))

def import_to_db(db_path, json_path):
    os.unlink(os.path.expanduser(db_path))
    session = get_session(db_path)
    cards = json.load(open(os.path.expanduser(json_path), 'r'))
    for card in cards:
        c = Card.from_dict(card)
        session.add(c)
    session.commit()





# -*- coding: utf-8 -*-

# Copyright (C) 2019 github.com/shyal
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from urwid import *

from vulcan.lib.model import VulcanModel
from vulcan.lib.prefs import prefs
from vulcan.lib.view import VulcanView

"""
Since we're trying to use the MVC, communication between the Model and the View
should happen through here. The bulk of the logic should also be implemented here.
"""


class GraphModel:
    def get_data(self, card):
        ret = []
        val = False
        if card:
            for i, h in enumerate(card.history):
                val += int(h.success)
                ret.append([i, val])
        return ret


class EditCard(WidgetWrap):

    def __init__(self, controller, card=None):
        self.card = card
        self.controller = controller
        self.preamble = Edit(caption="Preamble:\n", multiline=True, edit_text=(card.preamble if card else ''))
        self.question = Edit(caption="Question:\n", multiline=True, edit_text=(card.question if card else ''))
        self.answer =   Edit(caption="Answer:\n", multiline=True, edit_text=(card.answer if card else ''))
        self.tag = Edit(caption="Tag:\n", edit_text=(card.context if card else ''))
        ok = Button(label="Edit" if card else "Add")
        cancel = Button(label="Return")
        connect_signal(ok, 'click', self.ok_pressed)
        connect_signal(cancel, 'click',
            lambda button:controller.cancel_card_inserted())

        line = Divider(u'_')

        lw = SimpleFocusListWalker([
            self.preamble,
            AttrMap(line, 'line'),
            self.question,
            AttrMap(line, 'line'),
            self.answer,
            AttrMap(line, 'line'),
            self.tag,
            AttrMap(line, 'line'),
            ok,
            cancel
            ])

        lb = ListBox(lw)
        super(EditCard, self).__init__(lb)

    def reset_fields(self):
        self.question.set_edit_text('')
        self.answer.set_edit_text('')
        self.preamble.set_edit_text('')
        self.tag.set_edit_text('')

    def ok_pressed(self, button):
        if self.card:
            self.card.question = self.question.edit_text
            self.card.answer = self.answer.edit_text
            self.card.preamble = self.preamble.edit_text
            self.card.context = self.tag.edit_text
            self.controller.commit()
            self.controller.cancel_card_inserted()
        else:
            self.controller.new_card_inserted(
                question=self.question.edit_text,
                answer=self.answer.edit_text,
                preamble=self.preamble.edit_text,
                tag=self.tag.edit_text)
            self.reset_fields()


class VulcanController:
    """
    A class responsible for setting up the model and view and running
    the application.
    """
    def __init__(self, db_path):
        self.__card = None
        self.__model = VulcanModel(db_path)
        self.graph_model = GraphModel()
        self.view = VulcanView(controller=self)

    def main(self, screenshot_mode=False):
        palette = [
            ('heading', 'black', 'light gray'),
            ('line', 'black', ''),
            ('options', 'dark gray', 'black'),
            ('focus heading', 'white', 'dark red'),
            ('focus line', 'black', 'dark red'),
            ('focus options', 'black', 'light gray'),
            ('dark blue', 'dark blue', '', 'standout'),
            ('dark green', 'dark green', '', 'standout'),
            ('brown', 'brown', '', 'standout'),
            ('bg 1',         'black',      'dark blue', 'standout'),
            ('bg 1 smooth',  'dark blue',  'black'),
            ('bg 2',         'black',      'dark cyan', 'standout'),
            ('bg 2 smooth',  'dark cyan',  'black'),
            ('bg 1 smooth',  'dark blue',  'black'),
            ('bg background','black', 'black'),
        ]

        focus_map = {
            'heading': 'focus heading',
            'options': 'focus options',
            'line': 'focus line'}

        if screenshot_mode:
            palette.extend([
                ('line', 'white', 'light gray'),
                ('', 'white', 'light gray'),
                (None, 'white', 'light gray'),
            ])

        self.loop = MainLoop(widget=self.view, palette=palette, unhandled_input=self.unhandled_input)
        self.loop.run()

    def set_tag(self, tag, state):
        prefs["tag_"+tag] = state
        self.__model.tags[tag] = state
        self.__model.init_algo()

    def get_graph_data(self):
        return self.graph_model.get_data(self.__card)

    def draw_card(self):
        self.__card = self.__model.draw_card()
        if hasattr(self, 'view'):
            self.view.update_graph()
        return self.__card

    def reply(self, *args, **kwargs):
        self.__model.algo.reply(*args, **kwargs)

    def get_tags(self):
        return self.__model.tags.items()

    def commit(self):
        self.__model.algo.session.commit()

    def new_card_inserted(self, question, answer, tag='', preamble=''):
        self.__model.add_card(question=question, answer=answer, context=tag, preamble=preamble)
        self.__model.refresh_tags()
        self.view.footer.set_text("Card added successfully")
        self.view.refresh_tags()
        self.view.update_graph()

    def card_edited(self):
        self.view.columns.widget_list[0] = self.view.term
        self.__model.refresh_tags()
        self.view.refresh_tags()
        self.view.columns.widget_list.append(self.view.sidebar)
        self.view.update_graph()

    def cancel_card_inserted(self):
        # self.view.sidebar = self.view.tag_controls()
        self.view.columns.widget_list.append(self.view.sidebar)
        self.view.columns.widget_list[0] = self.view.term
        self.__model.refresh_tags()
        self.view.refresh_tags()
        self.view.terminal.draw_card()
        self.view.update_graph()

    def unhandled_input(self, key):
        """
        Handle Keyboard Input Here
        """
        if key == 'ctrl n':
            # On New Card
            self.view.columns.widget_list.pop()
            self.view.columns.widget_list[0] = EditCard(self)
        elif key == 'ctrl e':
            # On Edit Card
            self.view.columns.widget_list.pop()
            self.view.columns.widget_list[0] = EditCard(self, card=self.__card)
        elif key == 'ctrl p':
            # On Peek
            self.view.terminal.peek()
        elif key == 'f1':
            self.view.open_help_dialog()
        elif key == 'f3':
            self.view.open_grid_dialog()
        elif key == 'f5':
            self.view.open_stats_dialog()
        elif key == 'ctrl b':
            self.__model.pause_card(self.__card)
            self.view.terminal.card = None
            self.__model.refresh_tags()
            self.view.columns.widget_list[2] = self.view.sidebar
            self.view.terminal.draw_card()
        elif key == 'ctrl d':
            # On Card Deletion
            if self.view.terminal.card:
                self.__model.remove_card(self.view.terminal.card)
                self.view.terminal.card = None
                self.__model.refresh_tags()
                self.view.columns.widget_list[2] = self.view.sidebar
                self.view.terminal.draw_card()



# -*- coding: utf-8 -*-

# Copyright (C) 2019 github.com/shyal
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from functools import partial
from urwid import Edit, Text

import vulcan.lib.context as ctx


def wrong_cb(given_answer, right_answer, lw):
    lw.append(Text("wrong.. you wrote:"))
    lw.append(Text(given_answer))
    lw.append(Text("please remember:"))
    lw.append(Text(right_answer))


class VulcanTerminal(Edit):

    """
    This class may be totally overcomplicated.
    """

    def __init__(self, lw, controller, view, *args, **kwargs):
        self.lw = lw
        self.controller = controller
        self.view = view
        self.card = None
        super(VulcanTerminal, self).__init__(*args, **kwargs)
        self.draw_card()

    def draw_card(self):
        new_card = self.controller.draw_card()
        if new_card:
            if self.card != new_card:
                self.card = new_card
                try:
                    self.lw.remove(self)
                except:
                    pass
                self.card_context = ctx.prompt(self.card)
                if self.card_context.command:
                    self.lw.append(Text(ctx.bash_prompt() + self.card_context.command))
                self.lw.append(Text(self.card_context.render()))
                self.lw.append(self)
                self.lw.set_focus(len(self.lw) - 1)
                self.set_caption(self.card_context.prompt)
                self.view.update_graph()
        else:
            self.view.footer.set_text("No Cards - press ctrl-n to insert a new one")

    def peek(self):
        try:
            self.lw.remove(self)
        except:
            pass
        self.lw.append(Text(self.card.answer))
        self.lw.set_focus(len(self.lw) - 1)
        self.set_caption(self.card_context.prompt)

    def keypress(self, size, key):
        fin_multi = self.card.multi and len(self.edit_text) and key == 'enter' and self.edit_text[-1] == '\n'
        fin_norm =  ((not self.card.multi) and key == 'enter')
        if (not (fin_norm or fin_multi)):
            return super(VulcanTerminal, self).keypress(size, key)
        self.lw.append(Text(self.card_context.prompt + [self.edit_text]))
        self.controller.reply(self.card.id, self.card.compare_answer(self.edit_text), self.edit_text, partial(wrong_cb, lw=self.lw))
        self.draw_card()
        self.set_edit_text('')
        self.lw.set_focus(len(self.lw)-1)


# -*- coding: utf-8 -*-

# Copyright (C) 2017 github.com/shyal
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import division
import math
from vulcan.lib.store import *
from sqlalchemy import and_, or_
import random
import operator
import time

def stats(session):
    query = session.query(Card)
    num = query.count()
    new = query.filter(Card.successes == 0).count()
    paused = query.filter(Card.paused == True).count()
    young = query.filter(and_(Card.paused == False, Card.successes <= 7)).count()
    mid = query.filter(and_(Card.paused == False, Card.successes > 7, Card.successes <= 9)).count()
    mature = query.filter(and_(Card.paused == False, Card.successes > 9)).count()
    return "Cards: {0}\nNew: {1}\nPaused: {2}\nYoung: {3}\nMid: {4}\nMature: {5}".format(num, new, paused, young, mid, mature)

def weighted_pick(d):
    r = random.uniform(0, sum(d.itervalues()))
    s = 0.0
    for k, w in d.iteritems():
        s += w
        if r < s: return k
    return k

def simple_norm(D):
    newd = {}
    S = sum(D.values())
    N = len(D.keys())
    if S == 0:
        for c in D.keys():
            newd[c] = float(1)/N
    else:
        for c in D.keys():
            newd[c] = float(D[c])/S
    return newd

def get_telapsed(last_draw_timestamp):
    telapsed = {}
    now = datetime.now()
    for card_id, tstamp in last_draw_timestamp.items():
        telapsed[card_id] = (now - tstamp).seconds
    return telapsed

def get_expectation_of_recall(tdist, strength):
    E_recall = {}
    for card_id in tdist.keys():
        t = tdist[card_id]
        s = strength[card_id]
        E_recall[card_id] = math.exp(-t/s)
    return E_recall

def weighted_combination(weights, distributions):
    sum_weights = sum(weights)
    newd = {}
    ndist = len(distributions)
    for card_id in distributions[0]:
        weighted_sum = 0
        for i in range(ndist):
            weighted_sum += (weights[i] * distributions[i][card_id])
        newd[card_id] = weighted_sum / sum_weights
    return newd

def laplace_smoothing_dist(ncards_drawn, card_count):
    dist = {}
    nwords = len(card_count)
    for card_id in card_count:
        dist[card_id] = (float(card_count[card_id]) + 1)/ (ncards_drawn + nwords)
    dist = simple_norm(dist)
    return dist

class FlashcardAlgorithm():
    def __init__(self, wdist_weight=1, tdist_weight=0, includeMuli=1, includePaused=0, tags=[]):
        self.session = get_session()
        self.tags = tags
        self.wdist_weight = wdist_weight
        self.tdist_weight = tdist_weight
        self.includePaused = includePaused
        self.includeMuli = includeMuli
        self.init()

    def init(self):
        filt = and_(or_(Card.paused == 0, Card.paused == self.includePaused), or_(Card.multi == 0, Card.multi == self.includeMuli))
        self.learning_rate = 100
        self.draw_dist = {}
        self.strength = {}
        self.weakness = {}
        self.last_draw_timestamp = {}
        self.unknown_count = {}
        self.known_count = {}
        self.ncards_drawn = 0
        if self.tags:
            filt = and_(filt, Card.context.in_(self.tags))
        for card in self.session.query(Card).filter(filt):
            self.ncards_drawn += card.attempts
            self.known_count[card.id] = card.successes
            self.unknown_count[card.id] = card.failures
            self.last_draw_timestamp[card.id] = card.last_draw_timestamp
        self.strength = laplace_smoothing_dist(self.ncards_drawn, self.known_count)
        self.weakness = laplace_smoothing_dist(self.ncards_drawn, self.unknown_count)

    def get_tdist(self):
        telapsed = get_telapsed(self.last_draw_timestamp)
        tdist = simple_norm(telapsed)
        return tdist

    def show_scores(self, _id = None):
        tdist = self.get_tdist()
        E_recall = get_expectation_of_recall(tdist, self.strength)
        if _id:
            return E_recall[_id]
        else:
            return sorted(E_recall.items(), key=operator.itemgetter(1))

    def show_strength(self, _id=None):
        if _id:
            return self.strength[_id]
        else:
            return sorted(self.strength.items(), key=operator.itemgetter(1), reverse=True)

    def draw_card(self):
        tdist = self.get_tdist()
        E_recall = get_expectation_of_recall(tdist, self.strength)
        self.draw_dist = weighted_combination([self.wdist_weight, self.tdist_weight], [self.weakness, tdist])
        sorted_cards = sorted(E_recall.items(), key=operator.itemgetter(1))
        if sorted_cards:
            card_id = sorted_cards[0][0]
            self.last_draw_timestamp[card_id] = datetime.now()
            self.ncards_drawn += 1
            return self.session.query(Card).filter(Card.id == card_id).first()

    def reply(self, card_id, correct, given_answer, wrong_cb):
        card = self.session.query(Card).filter(Card.id == card_id).first()
        now = int(time.time())
        if correct:
            self.known_count[card_id] += 1
        else:
            self.unknown_count[card_id] += 1
            if wrong_cb:
                wrong_cb(given_answer=given_answer, right_answer=str(card.answer))
        card.history.append(History(success=correct, date=now))
        self.session.commit()
        return True



#!/usr/bin/env python

try:
    from setuptools.core import setup
except ImportError:
    from distutils.core import setup

from os import path
import os

this_directory = path.abspath(path.dirname(__file__))
readme_path = path.join(this_directory, 'README.md')
if os.path.exists(readme_path):
    with open(readme_path) as f:
        long_description = f.read()
else:
    long_description = ""

setup(name='vulcan',
      version='0.1.12',
      description='Terminal-based flashcard application, for developers, that uses machine learning to schedule reviews',
      author='Shyal',
      author_email='shyal@shyal.com',
      url='https://www.github.com/shyal/vulcan/',
      packages=['vulcan', 'vulcan.lib', 'vulcan.vulcan_db', 'vulcan.vulcan_db.versions'],
      scripts=['vulcan/vulcan'],
      long_description=long_description,
      long_description_content_type='text/markdown',
      install_requires=[
      'urwid',
      'pygments',
      'sqlalchemy',
      'arrow',
      'sqlalchemy-migrate',
      ])




# -*- coding: utf-8 -*-

# Copyright (C) 2019 github.com/shyal
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


from textwrap import dedent
from urwid import *
import vulcan.lib.utils
from vulcan.lib.store import get_session, Card
from vulcan.lib.prefs import prefs
from vulcan.lib.terminal import VulcanTerminal


class VulcanView(WidgetPlaceholder):
    """
    A class responsible for providing the application's interface and
    graph display.
    """
    def __init__(self, controller):
        self.controller = controller
        super(VulcanView, self).__init__(SolidFill(u'/'))

        self.main_window()

    def on_mode_button(self, button, state):
        """Notify the controller of a new mode setting."""
        self.controller.set_tag(button.get_label(), state)
        self.terminal.draw_card()

    def exit_program(self, w=None):
        prefs.close()
        raise ExitMainLoop()

    def refresh_tags(self):
        """
        Loop through the tags available from the model
        and either update, or add them to the list walker
        """
        for k, v in self.controller.get_tags():
            cb = list(filter(lambda x: x.get_label() == k, self.lw))
            if len(cb):
                cb = cb[0]
                cb.set_state(v)
            else:
                self.lw.append(CheckBox(label=k, state=v, on_state_change=self.on_mode_button))

        # we also need to handle removal..
        model_tags = set([k for k, v in self.controller.get_tags()])
        list_walker_tags = set([x.get_label() for x in self.lw])
        to_remove = list_walker_tags - model_tags
        for tr in to_remove:
            self.lw.remove(tr)

    def bar_graph(self):
        return BarGraph(['bg background','bg 1','bg 1'])

    def update_graph(self, force_update=False):
        l = self.controller.get_graph_data()
        if len(l) > 2:
            self.graph.set_data(l, max([x[1] for x in l]) + 1)
        else:
            self.graph.set_data([], 0)
        return True

    def tag_controls(self):
        self.lw = SimpleListWalker([])
        self.refresh_tags()
        self.graph = self.bar_graph()
        self.graph_wrap = WidgetWrap( self.graph )
        self.update_graph()
        p = Pile([
            ('fixed', 1, ListBox([Text("Tags",align="center")])),
            ('weight', 1, ListBox(self.lw)),
            ('weight', 1, self.graph_wrap),
            ('fixed', 1, ListBox([Button("Quit", self.exit_program)]))
            ])
        return p

    def ok_pressed(self, _):
        self.original_widget = self.the_first


    def main_window(self):
        lw = SimpleListWalker([])
        self.term = ListBox(lw)
        self.vline = AttrWrap(SolidFill(u'\u2502'), 'line')
        self.sidebar = self.tag_controls()

        self.columns = Columns([
            ('weight', 3, self.term),
            ('fixed', 1, self.vline), ('weight', 1, self.sidebar)],
            dividechars=1, focus_column=0)

        self.header = AttrWrap(Columns([
            ('weight', 1, Text('vulcan 1.0')),
            ('weight', 1, Text('f1:help')),
            ('weight', 1, Text('f3:grid')),
            ('weight', 1, Text('f5:stats')),
        ]), 'bg 1 smooth')
        self.footer = AttrWrap(Text(""), 'bg 1 smooth')

        # the terminal should exist in the controller, not in the view
        self.terminal = VulcanTerminal(lw=lw, controller=self.controller, view=self, multiline=True)

        frame = Frame(body=self.columns, footer=self.footer, header=self.header)

        self.original_widget = Overlay(frame, self.original_widget, align='center', width=('relative', 100),
            valign='middle', height=('relative', 100),
            min_width=24, min_height=8)

        self.the_first = self.original_widget

    def open_help_dialog(self):

        ok = Button('Close')
        connect_signal(ok, 'click', self.ok_pressed)

        nested = AttrWrap(LineBox(ListBox(
            [
                Text(dedent("""
                    Vulcan: remember anything.

                    RTVM: read the vulcan manual at https://vulcan.sh

                    Keyboard Shortcuts:

                    ?           -       open this help menu
                    ctrl n      -       open add card menu
                    ctrl e      -       edit current card
                    ctrl d      -       delete current card

                    """)),
                Text(dedent("""
                                                        _____
                                               __...---'-----`---...__
                                          _===============================
                         ______________,/'      `---..._______...---'
                        (____________LL). .    ,--'
                         /    /.---'       `. /
                        '--------_  - - - - _/
                                  `~~~~~~~~'

                    """)),
                ok
            ]
        )), 'bg 1')

        self.original_widget = Overlay(nested, self.original_widget, align='center', width=('relative', 70),
            valign='middle', height=('relative', 70),
            min_width=10, min_height=8)


    def open_stats_dialog(self):

        ok = Button('Close')
        connect_signal(ok, 'click', self.ok_pressed)
        
        session = get_session()

        query = session.query(Card)

        nested = AttrWrap(LineBox(ListBox(
            [
                Text(dedent(f"""
                    Vulcan stats.

                    {utils.stats(total=True)}
                    {utils.stats(today=True)}

                    questions asked: {sum([x.attempts for x in query])}
                    successes: {sum([x.successes for x in query])}
                    failures: {sum([x.failures for x in query])}

                    """)),
                ok
            ]
        )), 'bg 1')

        self.original_widget = Overlay(nested, self.original_widget, align='center', width=('relative', 70),
            valign='middle', height=('relative', 70),
            min_width=10, min_height=8)



    def open_grid_dialog(self):

        ok = Button('Close')
        connect_signal(ok, 'click', self.ok_pressed)
        
        session = get_session()

        query = session.query(Card)

        w = {
            'ID': 5,
            'Tag': 10,
        }

        cards = [Columns([
                ('fixed', w['ID'], Text('ID')),
                ('fixed', 3, Text(' | ')),
                ('fixed', w['Tag'], Text('Tag')),
                ('fixed', 3, Text(' | ')),
                ('weight', 1, Text('Question')),
                ('fixed', 3, Text(' | ')),
                ('weight', 1, Text('Answer')),
                ('fixed', 3, Text(' | ')),
                ('weight', 1, Text('CSuccesses')),
                ('fixed', 3, Text(' | ')),
                ('weight', 1, Text('CFailures')),
                ('fixed', 3, Text(' | ')),
            ])]

        for card in query:
            cards.append(Columns([
                ('fixed', w['ID'], Text(str(card.id))),
                ('fixed', 3, Text(' | ')),
                ('fixed', w['Tag'], Text(str(card.context[:20]))),
                ('fixed', 3, Text(' | ')),
                ('weight', 1, Text(str(card.question[:20]))),
                ('fixed', 3, Text(' | ')),
                ('weight', 1, Text(str(card.answer[:20]))),
                ('fixed', 3, Text(' | ')),
                ('weight', 1, Text(str(card.successes))),
                ('fixed', 3, Text(' | ')),
                ('weight', 1, Text(str(card.failures))),
                ('fixed', 3, Text(' | ')),
            ]))

        cards.append(ok)

        nested = AttrWrap(LineBox(ListBox(cards)), 'bg 1')

        self.original_widget = Overlay(nested, self.original_widget, align='center', width=('relative', 90),
            valign='middle', height=('relative', 90),
            min_width=10, min_height=8)



